/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#ifdef _WIN32
#include "stdafx.h"
#endif
#include "include/com_hcrest_jfreespace_LibFreespace.h"
#include "freespace/freespace.h"
#include "freespace/freespace_codecs.h"

#ifdef __cplusplus
extern "C" {
#endif

JavaVM* jvm_ = NULL;
JNIEnv* env_ = NULL;
jclass  libfreespaceClass_;
jobject libfreespaceInstance_;
jmethodID deviceInsertedId_ = NULL;
jmethodID deviceRemovedId_ = NULL;
jmethodID readCallbackId_ = NULL;
jmethodID sendCallbackId_ = NULL;
jmethodID libfreespacePerformId_ = NULL;
jclass    deviceClass_;
jmethodID deviceConstructor_ = NULL;
HANDLE    waitHandles_[MAXIMUM_WAIT_OBJECTS];
DWORD     waitHandleCount_ = 0;
#define WAIT_MS_DEFAULT 1000
DWORD     waitMs_ = WAIT_MS_DEFAULT;

HANDLE javaEvent;

static void add_pollfd(FreespaceFileHandleType fd, short events) {
    DWORD i;

    for (i = 0; i < waitHandleCount_; i++) {
        if (waitHandles_[i] == fd) {
            return;
        }
    }

    if (waitHandleCount_ >= MAXIMUM_WAIT_OBJECTS) {
        // If there are too many handles, then wait on the ones we can, and set a 
        // 1 ms timeout so at least we call freespace_perform() at a reasonable interval.
        waitMs_ = 1;  
        return;
    }

    waitMs_ = WAIT_MS_DEFAULT;
    waitHandles_[waitHandleCount_] = fd;
    waitHandleCount_++;

    printf("jfreespaceNative: Added a handle %p. Total handles = %d\n", fd, waitHandleCount_);
}

static void remove_pollfd(FreespaceFileHandleType fd) {
    DWORD i;

    for (i = 0; i < waitHandleCount_; i++) {
        if (waitHandles_[i] == fd) {
            break;
        }
    }

    if (i != waitHandleCount_) {
        memmove(&waitHandles_[i], &waitHandles_[i + 1], (waitHandleCount_ - i - 1) * sizeof(HANDLE));
        waitHandleCount_--;
    }

    printf("jfreespaceNative: Removed a handle %p. Total handles = %d\n", fd, waitHandleCount_);
}



/** 
 * Handle any pending JVM exceptions
 * @param env The Java environment.
 */
void handleJvmExceptions(JNIEnv* env) {
    // Handle any exceptions thrown during the method call
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	}
}

/**
 * Callback for handling device insertion and device removal events.
 * 
 * @param event The hotplug event type: insertion or removal.
 * @param id The device affected by this event.
 * @param cookie The UNUSED data specified during configuration.
 */
void jfreespace_hotplugCallback(enum freespace_hotplugEvent event,
                                FreespaceDeviceId id,
                                void* cookie) {
    if (event == FREESPACE_HOTPLUG_INSERTION) {
        // Get the device information
        struct FreespaceDeviceInfo info;
        jstring retStr;
        jobject result;

        freespace_getDeviceInfo(id, &info);

        // Create the Device instance
        retStr = (*env_)->NewStringUTF(env_, info.name);
        handleJvmExceptions(env_);
        result = (*env_)->NewObject(env_, deviceClass_, deviceConstructor_, id, retStr, info.vendor, info.product, info.hVer);
        handleJvmExceptions(env_);

        // Pass the Device instance back to Java.
        (*env_)->CallVoidMethod(env_, libfreespaceInstance_, deviceInsertedId_, result);
        handleJvmExceptions(env_);

        // Clean up after ourselves.
        (*env_)->DeleteLocalRef(env_, retStr);
        (*env_)->DeleteLocalRef(env_, result);

    } else if (event == FREESPACE_HOTPLUG_REMOVAL) {
        (*env_)->CallVoidMethod(env_, libfreespaceInstance_, deviceRemovedId_, id);
        handleJvmExceptions(env_);
    } else {
        // Unknown hotplug event;
    }
}

/**
 * Callback for asynchronous message sent.
 * 
 * @param id The device to which the message was sent.
 * @param cookie The data (send message id) specified during the call to send.
 * @param result FREESPACE_SUCCESS if a packet was sent; else error code
 */
void jfreespace_sendCallback(FreespaceDeviceId id, void* cookie, int result) {
	// Note that there is the potential for data loss when going from cookie to jint. The casts below
	// prevent compiler warnings but cookie must still be used with care.
    (*env_)->CallVoidMethod(env_, libfreespaceInstance_, sendCallbackId_, id, result, (jint) ((uint64_t) cookie));
    handleJvmExceptions(env_);
}


/**
 * Callback for received Freespace events.
 *
 * @param id The device that generated the message
 * @param message the raw HID report
 * @param length its length
 * @param cookie the UNUSED data passed to freespace_setReceiveCallback().
 * @param result FREESPACE_SUCCESS if a packet was received; else error code
 */
void jfreespace_receiveCallback(FreespaceDeviceId id,
                                          const unsigned char* message,
                                          int length,
                                          void* cookie,
                                          int result) {
    // Construct the JVM message
    jbyteArray msg = (*env_)->NewByteArray(env_, length);
    if (msg == NULL) {
        // Most likely out of memory.  Not much we can do about it.
        handleJvmExceptions(env_);
        return;
    } else {
        jbyte* pMsg;
        int i;
		LARGE_INTEGER ts;
		BOOL tsErr;
		DWORD_PTR oldmask;

        pMsg = (*env_)->GetByteArrayElements(env_, msg, NULL);
        for (i = 0; i < length; i++) 
        {
            pMsg[i] = (jbyte) message[i];
        }
        (*env_)->ReleaseByteArrayElements(env_, msg, pMsg, 0);

		// Timestamp the message
		oldmask = SetThreadAffinityMask(GetCurrentThread(), 0);
		tsErr = QueryPerformanceCounter(&ts);
		SetThreadAffinityMask(GetCurrentThread(), oldmask);
		if (!tsErr) {
			ts.QuadPart = GetLastError();
		}

        // Send the message to the JVM
		(*env_)->CallVoidMethod(env_, libfreespaceInstance_, readCallbackId_, id, msg, result, ts.QuadPart);
        handleJvmExceptions(env_);

        // Release the message bytes
        (*env_)->DeleteLocalRef(env_, msg);
    }
}

/*
 * Class:     com_hcrest_jfreespace_LibFreespace
 * Method:    native_init
 * Signature: (Lcom/hcrest/jfreespace/LibFreespace;)I
 */
JNIEXPORT jint JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1init
    (JNIEnv * env, jclass clazz, jobject instance) {

    jclass localRefCls;
    int rc;

    env_ = env;

    // Initialize the Java references.
    (*env)->GetJavaVM(env, &jvm_);
	libfreespaceInstance_ = (jobject) (*env)->NewGlobalRef(env, instance);
    handleJvmExceptions(env);

	// Get the Discovery methods of interest
	libfreespaceClass_ = (jclass) (*env)->NewGlobalRef(env, (*env)->GetObjectClass(env, libfreespaceInstance_));
    if (libfreespaceClass_ == NULL) {
        return -1;
    }
    handleJvmExceptions(env);
    deviceInsertedId_   = (*env)->GetMethodID(env, libfreespaceClass_, "deviceInserted", "(Lcom/hcrest/jfreespace/Device;)V");
    handleJvmExceptions(env);
    deviceRemovedId_    = (*env)->GetMethodID(env, libfreespaceClass_, "deviceRemoved", "(I)V");
    handleJvmExceptions(env);
    readCallbackId_     = (*env)->GetMethodID(env, libfreespaceClass_, "readCallback", "(I[BIJ)V");
    handleJvmExceptions(env);
    sendCallbackId_     = (*env)->GetMethodID(env, libfreespaceClass_, "sendCallback", "(III)V");
    handleJvmExceptions(env);
    libfreespacePerformId_ = (*env)->GetMethodID(env, libfreespaceClass_, "perform", "()I");
    handleJvmExceptions(env);
    if (deviceInsertedId_ == NULL || deviceRemovedId_ == NULL || readCallbackId_ == NULL || 
        sendCallbackId_ == NULL || libfreespacePerformId_ == NULL) {
        return -1;
    }

    // Get the Device class
    localRefCls = (jclass) (*env)->FindClass(env, "com/hcrest/jfreespace/Device");
    if (localRefCls == NULL) {
        return -1;
    }
    deviceClass_ = (jclass) (*env)->NewGlobalRef(env, localRefCls);
    (*env)->DeleteLocalRef(env, localRefCls);
    handleJvmExceptions(env);

    // Get the Device constructor method
    deviceConstructor_  = (*env)->GetMethodID(env, deviceClass_, "<init>", "(ILjava/lang/String;III)V");
    handleJvmExceptions(env);

    // Initialize the Java event
    javaEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (javaEvent == NULL) { 
        printf("jfreespaceNative: CreateEvent failed (%d)\n", GetLastError());
    } else {
        add_pollfd(javaEvent, 1);
    }

    // Initialize callbacks
    rc = freespace_init();
    if (FREESPACE_SUCCESS != rc) {
        return rc;
    }
    freespace_setDeviceHotplugCallback(jfreespace_hotplugCallback, NULL);
    freespace_setFileDescriptorCallbacks(add_pollfd, remove_pollfd);
    freespace_syncFileDescriptors();

    return FREESPACE_SUCCESS;
}

/*
 * Class:     com_hcrest_jfreespace_LibFreespace
 * Method:    native_exit
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1exit
  (JNIEnv *env, jclass clazz) {
    // Clear the Java variables
	if (libfreespaceInstance_ != NULL) {
		(*env)->DeleteGlobalRef(env, libfreespaceInstance_);
		libfreespaceInstance_ = NULL;
	}

	if (libfreespaceClass_ != NULL) {
		(*env)->DeleteGlobalRef(env, libfreespaceClass_);
		libfreespaceClass_ = NULL;
	}

    if (deviceClass_ != NULL) {
		(*env)->DeleteGlobalRef(env, deviceClass_);
		deviceClass_ = NULL;
	}
    deviceInsertedId_ = NULL;
    deviceRemovedId_ = NULL;
    readCallbackId_ = NULL;
    libfreespacePerformId_ = NULL;
    deviceConstructor_ = NULL;

    remove_pollfd(javaEvent);
    CloseHandle(javaEvent);

    freespace_exit();
}


/*
 * Class:     com_hcrest_jfreespace_LibFreespace
 * Method:    native_openDevice
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1openDevice
  (JNIEnv *env, jclass clazz, jint id) {
    int rc;
    freespace_private_setReceiveCallback(id, jfreespace_receiveCallback, NULL);
    rc = freespace_openDevice(id);
    if (rc != FREESPACE_SUCCESS) {
        return rc;
    }
    rc = freespace_flush(id);
    if (rc != FREESPACE_SUCCESS) {
        return rc;
    }
    return FREESPACE_SUCCESS;
}

/*
 * Class:     com_hcrest_jfreespace_Device
 * Method:    native_closeDevice
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1closeDevice
(JNIEnv *env, jclass clazz, jint id) {
    freespace_closeDevice(id);
}

/*
 * Class:     com_hcrest_jfreespace_Device
 * Method:    native_send
 * Signature: (I[BI)I
 */
JNIEXPORT jint JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1send
(JNIEnv *env, jclass clazz, jint id, jbyteArray jmsg, jint cookie) {

    char hidMsg[FREESPACE_MAX_OUTPUT_MESSAGE_SIZE];
    int length;
    jbyte* pMsg;
    int i;

    // Get the message from Java
    pMsg = (*env)->GetByteArrayElements(env, jmsg, FALSE);
    if (pMsg == NULL) {
        return -10001;
    }

    // Copy into local buffer
    length = (*env)->GetArrayLength(env, jmsg);
    if (length > FREESPACE_MAX_OUTPUT_MESSAGE_SIZE) {
        return -10002;
    }
    for (i = 0; i < length; i++) {
        hidMsg[i] = pMsg[i];
    }
    (*env)->ReleaseByteArrayElements(env, jmsg, pMsg, 0);

    // Send the message
    return freespace_private_sendAsync(id, &hidMsg[0], length, 1000, jfreespace_sendCallback, (void*) ((uint64_t) cookie));
}

/*
 * Class:     com_hcrest_jfreespace_Device
 * Method:    native_flush
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1flush
(JNIEnv *env, jclass clazz, jint id) {
    return freespace_flush(id);
}

/*
 * Class:     com_hcrest_jfreespace_LibFreespace
 * Method:    native_signal
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1signal
(JNIEnv *env, jclass clazz) {
    SetEvent(javaEvent);
    return;
}

/*
 * Class:     com_hcrest_jfreespace_LibFreespace
 * Method:    native_loop
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_hcrest_jfreespace_LibFreespace_native_1loop
(JNIEnv *env, jclass clazz) {

    int quit;
    int rc;

    env_ = env;
    
    quit = FALSE;
    while (!quit) {
        // TODO : handle all the messages.
        DWORD status = WaitForMultipleObjects(waitHandleCount_, waitHandles_, FALSE, waitMs_);
        if (status == WAIT_FAILED) {
            Sleep(1); // Avoid tight loop if wait fails
        }

        freespace_perform();
        ResetEvent(javaEvent);
        rc = (*env_)->CallIntMethod(env_, libfreespaceInstance_, libfreespacePerformId_);
        if (rc < 0) {
            quit = TRUE;
        } else {
            handleJvmExceptions(env_);
        }
    }
}


#ifdef __cplusplus
}
#endif