/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import com.hcrest.jfreespace.inreport.DecoderFactory;
import com.hcrest.jfreespace.inreport.HidInMsg;
import com.hcrest.jfreespace.inreport.DecoderFactory.HID_MESSAGE_VERSION;
import com.hcrest.jfreespace.outreport.HidOutMsg;


/**
 * Provide complete access to a single Freespace device.
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
 * @version 1.0
 */
public class Device implements DeviceAccessInterface {
    private final int id_;
    private DeviceListenerInterface listener_ = null;
    //private static Logger log_ = Logger.getLogger(Device.class.getName()); 
    
    private final String name_;
    private final int vendorId_;
    private final int productId_;
    private final DecoderFactory.HID_MESSAGE_VERSION hVer_;
    private DeviceStatistics stats_ = null;
    
    private long sendMsgCnt_ = 0;
    private long sendMsgBytes_ = 0;
    private long sendMsgError_ = 0;
    private long readMsgCnt_ = 0;
    private long readMsgBytes_ = 0;
    private long readMsgError_ = 0;

    /**
     * Create a new device.
     * @param id The device handle.
     * @param name The user-meaningful name.
     * @param vendorId The vendor ID.
     * @param productId The product ID.
     */
    public Device(int id, String name, int vendorId, int productId, int hVer) {
        id_ = id;
        name_ = name;
        vendorId_ = vendorId;
        productId_ = productId;
        hVer_ = HID_MESSAGE_VERSION.values()[hVer];
        
        resetStatistics();
    }

    /**
     * @return The HID version.
     */
    public HID_MESSAGE_VERSION getHidVer() {
        return hVer_;
    }
    
    /**
     * @return The device ID.
     */
    int getId() {
        return id_;
    }
    
    /**
     * Get the vendor ID for this device.
     * @return The vendor ID.
     */
    public int getVendorId() {
        return vendorId_;
    }
    
    /**
     * Get the product ID for this device.
     * @return The product ID.
     */
    public int getProductId() {
        return productId_;
    }

    /**
     * Read callback from libfreespace.
     * @param msg The message received.
     * @param rc The return code from the received message.
     * @param ts A relative timestamp indicating when the message was received.
     */
    void readCallback(byte[] msg, int rc, long ts) {
        if (rc == FreespaceErrorCodes.FREESPACE_ERROR_NOT_FOUND.getErrorCode()) {
            readMsgError_++;
            LibFreespace.getInstance().deviceRemoved(id_);
            return;
        }
        if (msg == null) {
            readMsgError_++;
            return;
        }
        if (rc != FreespaceErrorCodes.FREESPACE_SUCCESS.getErrorCode()) {
            readMsgError_++;
        }
        readMsgBytes_ += msg.length;
        readMsgCnt_++;
        if (listener_ != null) {
            HidInMsg m = null;
            try {
                m = DecoderFactory.getDecoder(hVer_).decode(msg);
            } catch (Exception e) {
                // If the selected decode failed and if it was FSP2xx then try an FSP16x decode
                if (hVer_ == HID_MESSAGE_VERSION.FSP2xx) {
                    try {
                        m = DecoderFactory.getDecoder(HID_MESSAGE_VERSION.FSP16x).decode(msg);
                    } catch (Exception e2) {
                        return;
                    }
                }
            }
            listener_.handleRead(this, m, rc, ts);
        }
    }
    
    /**
     * Send callback from libfreespace.
     * @param msg The message received.
     * @param rc The return code from the received message.
     */
    void sendCallback(FreespaceEvent ev) {
        if (ev == null) {
            return;
        }
        sendMsgBytes_ += ev.getMessage().getMsg(hVer_).length;
        sendMsgCnt_++;

        if (ev.getReturnCode() != FreespaceErrorCodes.FREESPACE_SUCCESS.getErrorCode()) {
            sendMsgError_++;
        }
        if (listener_ != null) {
            listener_.handleSend(this, ev.getMessage(), ev.getReturnCode());
        }
    }
    
    /**
     * Indicate that this device has been removed from the system.
     * Close the device and notify all observers.
     */
    void removed() {
        if (listener_ != null) {
            listener_.notifyRemoved(this);
        }
        close();
    }
    
    public FreespaceErrorCodes open(DeviceListenerInterface listener) {
        listener_ = listener;
        resetStatistics();
        FreespaceEvent ev = new FreespaceEvent(FreespaceEvent.Type.OPEN, id_);
        int rc = LibFreespace.sendEventSynchronized(ev);
        return FreespaceErrorCodes.getInstance(rc);
    }
    
    public void close() {
        FreespaceEvent ev = new FreespaceEvent(FreespaceEvent.Type.CLOSE, id_);
        LibFreespace.sendEventSynchronized(ev);
        listener_ = null;
    }

    public FreespaceErrorCodes flush() {
        FreespaceEvent ev = new FreespaceEvent(FreespaceEvent.Type.FLUSH, id_);
        int rc = LibFreespace.sendEventSynchronized(ev);
        return FreespaceErrorCodes.getInstance(rc);
    }
    
    public FreespaceErrorCodes sendMessage(HidOutMsg msg) {
        byte[] msgBytes = msg.getMsg(hVer_);
        if (msgBytes == null) {
            return FreespaceErrorCodes.FREESPACE_ERROR_COULD_NOT_ENCODE;
        }
        sendMsgBytes_ += msgBytes.length;
        sendMsgCnt_++;
        FreespaceEvent ev = new FreespaceEvent(FreespaceEvent.Type.SEND, id_, msg, msgBytes);
        int rc = LibFreespace.sendEventSynchronized(ev);
        return FreespaceErrorCodes.getInstance(rc);
    }
    
    public FreespaceErrorCodes sendMessageAsync(HidOutMsg msg) {
        byte[] msgBytes = msg.getMsg(hVer_);
        if (msgBytes == null) {
            return FreespaceErrorCodes.FREESPACE_ERROR_COULD_NOT_ENCODE;
        }
        FreespaceEvent ev = new FreespaceEvent(FreespaceEvent.Type.SEND, id_, msg, msgBytes);
        LibFreespace.sendEvent(ev);
        return FreespaceErrorCodes.FREESPACE_SUCCESS;
    }
    
    /**
     * Returns a detailed string representation of the object.
     * @return A multiple line string representation of the object.
     */
    public String toDetailedString() {
        StringBuffer buf = new StringBuffer();
        buf.append(this.toString() + "\n");
        buf.append("   Vendor ID  = 0x" + Integer.toHexString(vendorId_) + " (" + vendorId_ + ")\n");
        buf.append("   Product ID = 0x" + Integer.toHexString(productId_) + " (" + productId_ + ")");
        return buf.toString();
    }
    
    
    private void resetStatistics() {
        sendMsgCnt_ = 0;
        sendMsgBytes_ = 0;
        sendMsgError_ = 0;
        readMsgCnt_ = 0;
        readMsgBytes_ = 0;
        readMsgError_ = 0;
    }
    
    /**
     * Get the statistics for this device.
     * @param previous The previous statistics to use for comparison.  For example,
     * the statistics from one second ago can be used to determine the current
     * throughput.
     * @return The device statistics.
     */
    public DeviceStatistics getStatistics(DeviceStatistics previous) {
        DeviceStatistics next = new DeviceStatistics(previous, 
                readMsgCnt_, readMsgBytes_, readMsgError_, 
                sendMsgCnt_, sendMsgBytes_, sendMsgError_);
        return next;
    }
    
    /**
     * Get the statistics for this device relative to the last call to this
     * method.  The last statistics instance is maintained internally.  If
     * multiple actors call this method in your application, consider using 
     * <code>getStatistics(DeviceStatistics)</code> instead.
     * @return The device statistics.
     */
    public DeviceStatistics getStatistics() {
        stats_ = getStatistics(stats_);
        return stats_;
    }
    
    /** 
     * Returns the user-meaningful device name.
     * This string is guaranteed to be a unique identifier for this device.
     * @return The string representation of this device instance.
     */
    @Override
    public String toString() {
        return name_ + " (" + id_ + ")";
    }

}
