/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import java.util.HashSet;

/**
 * Implement the default filter based upon vendor ID and product ID.
 * @author <a href="mailto:Matt.Liberty@hcrest.com">Matt Liberty</a>
 * @version 1.0
 */
public class DeviceFilter implements DeviceFilterInterface {
    private HashSet<String> filter_ = new HashSet<String>();

    /**
     * Create an instance of DeviceFilter that filters devices based upon
     * vendor ID and product ID pairs.
     * @param filter The filter string.  Each vendor ID and product ID pair 
     * must be surrounded in brackets [], and separated by
     * a ';'.  The pairs may be separated by other whitespace including
     * new lines. For example<br>  
     * [vendorId, productId]; [vendorId, productId, (additional ignored arguments)]; ...
     */
    public DeviceFilter(String filter) {
        if (filter == null) {
            return;
        }
        String internalFilter = filter.replaceAll("[ \\n\\r\\t]+", "");
        String[] searchList = internalFilter.split(";");
        for (int i = 0; i < searchList.length; i++) {
            String[] params = searchList[i].trim().substring(1, searchList[i].length() - 1).split("[,\\s]+");
            String key = getKey(params[0], params[1]);
            filter_.add(key);
        }
    }

    /** {@inheritDoc} */
    public boolean filterDevice(Device device) {
        String key = getKey("0x" + getString(device.getVendorId()),
                            "0x" + getString(device.getProductId()));
        return filter_.contains(key);
    }


    /**
     * Correctly format the values into a key.
     * @param vendorId  The vendor ID.
     * @param productId The product ID.
     * @return The key.
     */
    private String getKey(String vendorId, String productId) {
        return getString(getInteger(vendorId)) +
               getString(getInteger(productId));
    }

    /**
     * Get the string value of the integer.
     * @param i
     * @return
     */
    private String getString(int i) {
        String str = "0000" + Integer.toString(i, 16);
        return str.substring(str.length() - 4, str.length());
    }


    /**
     * Get the integer value.
     * @param str The string.
     * @return The integer equivalent, or 0 if null string.
     */
    private int getInteger(String str) {
        if (str == null) {
            return 0;
        }
        str = str.replace(" ", "");
        if (str.substring(0, 2).equals("0x")) {
            return Integer.parseInt(str.substring(2), 16);
        } else {
            return Integer.parseInt(str);
        }
    }
}
