/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import java.text.DecimalFormat;


/**
 * Container for a snapshot of the device statistics from one device. 
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
 * @version 1.0
 */
public class DeviceStatistics {

    private final long sendMsgCnt_;
    private final long sendMsgBytes_;
    private final long sendMsgError_;
    
    private final long readMsgCnt_;
    private final long readMsgBytes_;
    private final long readMsgError_;

    private final long dsendMsgCnt_;
    private final long dsendMsgBytes_;
    private final long dsendMsgError_;
    
    private final long dreadMsgCnt_;
    private final long dreadMsgBytes_;
    private final long dreadMsgError_;

    DeviceStatistics() {
        sendMsgCnt_     = 0;
        sendMsgBytes_   = 0;
        sendMsgError_   = 0;
        
        readMsgCnt_     = 0;
        readMsgBytes_   = 0;
        readMsgError_   = 0;

        dsendMsgCnt_    = 0;
        dsendMsgBytes_  = 0;
        dsendMsgError_  = 0;
        
        dreadMsgCnt_    = 0;
        dreadMsgBytes_  = 0;
        dreadMsgError_  = 0;
    }
    
    DeviceStatistics(final DeviceStatistics previous, 
            final long readMsgCnt, final long readMsgBytes, final long readMsgError,
            final long sendMsgCnt, final long sendMsgBytes, final long sendMsgError) {
        sendMsgCnt_     = sendMsgCnt;
        sendMsgBytes_   = sendMsgBytes;
        sendMsgError_   = sendMsgError;
        
        readMsgCnt_     = readMsgCnt;
        readMsgBytes_   = readMsgBytes;
        readMsgError_   = readMsgError;
        
        if (previous != null) {
            dsendMsgCnt_    = sendMsgCnt_   - previous.sendMsgCnt_;
            dsendMsgBytes_  = sendMsgBytes_ - previous.sendMsgBytes_;
            dsendMsgError_  = sendMsgError_ - previous.sendMsgError_;
            
            dreadMsgCnt_    = readMsgCnt_   - previous.readMsgCnt_;
            dreadMsgBytes_  = readMsgBytes_ - previous.readMsgBytes_;
            dreadMsgError_  = readMsgError_ - previous.readMsgError_;
            
        } else {
            dsendMsgCnt_    = sendMsgCnt_;
            dsendMsgBytes_  = sendMsgBytes_;
            dsendMsgError_  = sendMsgError_;
            
            dreadMsgCnt_    = readMsgCnt_;
            dreadMsgBytes_  = readMsgBytes_;
            dreadMsgError_  = readMsgError_;
        }
    }
    
    public long getSendMessageCount() {
        return sendMsgCnt_;
    }
    public long getSendMessageBytes() {
        return sendMsgBytes_;
    }
    public long getSendMessageError() {
        return sendMsgError_;
    }
    
    public long getReadMessageCount() {
        return readMsgCnt_;
    }
    public long getReadMessageBytes() {
        return readMsgBytes_;
    }
    public long getReadMessageError() {
        return readMsgError_;
    }

    public long getDeltaSendMessageCount() {
        return dsendMsgCnt_;
    }
    public long getDeltaSendMessageBytes() {
        return dsendMsgBytes_;
    }
    public long getDeltaSendMessageError() {
        return dsendMsgError_;
    }
       
    public long getDeltaReadMessageCount() {
        return dreadMsgCnt_;
    }
    public long getDeltaReadMessageBytes() {
        return dreadMsgBytes_;
    }
    public long getDeltaReadMessageError() {
        return dreadMsgError_;
    }
        
    private String numberFormat(long v) {
        String blankStr = "           ";
        DecimalFormat formatter = new DecimalFormat("###,###,###");
        String vStr = formatter.format(v);
        return blankStr.substring(0, blankStr.length() - vStr.length()) + vStr;
    }
    
    private String numberFormatd(long v) {
        String blankStr = "      ";
        DecimalFormat formatter = new DecimalFormat("####");
        String vStr = formatter.format(v);
        return blankStr.substring(0, blankStr.length() - vStr.length()) + vStr;
    }
    
    /** 
     * Returns a user-meaningful string representation of the statistics.
     * The string is intended to be displayed with a fixed-width font.
     * @return The user-meaningful string representation.
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("Statistic   dSend   dRead   TotalSend   TotalRead\n");
        buf.append("Messages   " + numberFormatd(dsendMsgCnt_) + "  " + 
                                  numberFormatd(dreadMsgCnt_) + " " + 
                                  numberFormat(sendMsgCnt_) + " " + 
                                  numberFormat(readMsgCnt_) + "\n");
        buf.append("Bytes      " + numberFormatd(dsendMsgBytes_) + "  " + 
                                  numberFormatd(dreadMsgBytes_) + " " + 
                                  numberFormat(sendMsgBytes_) + " " + 
                                  numberFormat(readMsgBytes_) + "\n");
        buf.append("Error      " + numberFormatd(dsendMsgError_) + "  " + 
                                  numberFormatd(dreadMsgError_) + " " + 
                                  numberFormat(sendMsgError_) + " " + 
                                  numberFormat(readMsgError_));
        return buf.toString();
    }
}
