/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * Singleton class that supports Freespace device discovery.
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
 * @version 1.0
 */
public final class Discovery {
    private Set<DiscoveryListenerInterface> listeners_ = new HashSet<DiscoveryListenerInterface>();
    private boolean isInitialized_ = false;
    
    private static Discovery instance_ = null;
    //private static Logger log_ = Logger.getLogger(Discovery.class.getName());
    
    /**
     * Create an instance of Discovery.
     */
    private Discovery() {
        // No action required.
    }
    
    /**
     * Get the one and only <code>Discovery</code> instance.
     * @return The <code>Discovery</code> instance.
     */
    public static Discovery getInstance() {
        if (instance_ == null) {
            instance_ = new Discovery();
        }
        return instance_;
    }
    
    /**
     * Initialize jfreespace and the underlying libfreespace.
     * Normally, the application should just call <code>addListener</code>
     * without calling this method explicitly.
     * @return true on success, false on failure.
     * @see #addListener
     */
    public boolean initialize() {
        if (isInitialized_) {
            return true;
        }
        isInitialized_ = true;
        LibFreespace.getInstance().setDiscovery(this);
        FreespaceEvent ev = new FreespaceEvent(FreespaceEvent.Type.INITIALIZE, 0);
        boolean success = LibFreespace.sendEventSynchronized(ev) == 0;
        if (success) {
            isInitialized_ = true;
        }
        return success;
    }
    
    /**
     * Finalize jfreespace and the underlying libfreespace.
     * Normally, the application should just call <code>removeListener</code>
     * without calling this method explicitly.
     * @see #removeListener
     */
    public void exit() {
        FreespaceEvent ev = new FreespaceEvent(FreespaceEvent.Type.EXIT, 0);
        LibFreespace.sendEventSynchronized(ev);
        isInitialized_ = false;
    }
    
    /**
     * Add a listener for device insertion and removal notifications.
     * Automatically call initialize if needed.  The 
     * <code>freespaceDeviceInserted</code> method will be called for every
     * device already present in the system when the listener is added.
     * @param listener The listener.
     */
    public void addListener(DiscoveryListenerInterface listener) {
        listeners_.add(listener);
        if (!isInitialized_) {
            initialize();
        } else {
            // Send every installed device to the listener.
            for (Device device : getDevices()) {
                deviceInserted(device);
            }
        }
    }
    
    /**
     * Remove a listener for device insertion and removal notifications.
     * Automatically call exit if no listeners are registered.
     * @param listener The listener.
     */
    public void removeListener(DiscoveryListenerInterface listener) {
        listeners_.remove(listener);
        if (listeners_.size() <= 0) {
            exit();
        }
    }

    /**
     * Callback from libfreespace on device inserted.
     * @param device The inserted device.
     */
    void deviceInserted(Device device) {
        // Call all listeners.
        for (DiscoveryListenerInterface listener : listeners_) {
            listener.freespaceDeviceInserted(device);
        }
    }
    
    /**
     * Callback from libfreespace on device removed.
     * @param id The removed device id.
     */
    void deviceRemoved(Device device) {
        for (DiscoveryListenerInterface listener : listeners_) {
            listener.freespaceDeviceRemoved(device);
        }
    }
    
    /**
     * Get all currently connected devices.
     * @return The collection of devices.
     */
    public Collection<Device> getDevices() {
        return LibFreespace.getInstance().getDevices();
    }
    
    /**
     * Get the currently connected devices matching the specified filter.
     * @return The collection of devices.
     */
    public Collection<Device> getDevices(DeviceFilterInterface filter) {
        Collection<Device> allDevices = LibFreespace.getInstance().getDevices();
        ArrayList<Device> filteredDevices = new ArrayList<Device>();
        Iterator<Device> iter = allDevices.iterator();
        while (iter.hasNext()) {
            Device device = iter.next();
            if (filter != null && filter.filterDevice(device)) {
                filteredDevices.add(device);
            }
        }

        return filteredDevices;
    }

}
