/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import com.hcrest.jfreespace.outreport.HidOutMsg;

/**
 * A Freespace command event to the native libfreespace code.  This class
 * allows for the commands to be synchronized to the single libfreespace thread. 
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
 * @version 1.0
 */
class FreespaceEvent {

    /**
     * The command type.
     * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
     * @version 1.0
     */
    enum Type {
        /** Initialize the native libfreespace library. */
        INITIALIZE,
        /** Exit the native libfreespace library. */
        EXIT,
        /** Open the specified device. */
        OPEN,
        /** Close the specified device. */
        CLOSE,
        /** Flush the specified device. */
        FLUSH,
        /** Send a message to the specified device. */
        SEND;
    }
    
    private final Type type_;
    private final int deviceId_;
    private final HidOutMsg msg_;
    private final byte[] msgBytes_;
    private int returnCode_ = -1;
    
    /**
     * Create an instance of FreespaceEvent.
     * @param type The event type.
     * @param deviceId The device (if applicable).
     */
    FreespaceEvent(Type type, int deviceId) {
        type_ = type;
        deviceId_ = deviceId;
        msg_ = null;
        msgBytes_ = null;
    }
    
    /**
     * Create an instance of FreespaceEvent.
     * @param type The event type.
     * @param deviceId The device (if applicable).
     * @param msg The message (for SEND only).
     * @param msgBytes The message bytes (for SEND only).
     */
    FreespaceEvent(Type type, int deviceId, HidOutMsg msg, byte[] msgBytes) {
        type_ = type;
        deviceId_ = deviceId;
        msg_ = msg;
        msgBytes_ = msgBytes;
    }
    
    /**
     * @return The Freespace event type.
     */
    Type getType() {
        return type_;
    }
    
    /**
     * @return The device.
     */
    int getDeviceId() {
        return deviceId_;
    }
    
    /**
     * @return The message (for SEND only).
     */
    HidOutMsg getMessage() {
        return msg_;
    }
    
    /**
     * @return The message bytes (for SEND only).
     */
    byte[] getMessageBytes() {
        return msgBytes_;
    }
    
    /**
     * @return The return code from the event.
     */
    int getReturnCode() {
        return returnCode_;
    }
    
    /**
     * Set the return code and notify all observers.
     * @param rc The return code.
     */
    synchronized void setReturnCode(int rc) {
        returnCode_ = rc;
        notifyAll();
    }
    
    /**
     * Wait for notification that this event has completed.
     * @return True on successful completion, false on failure or timeout.
     */
    synchronized boolean waitBoolean(int timeout) {
        try {
            wait(timeout);
        } catch (InterruptedException e) {
            return false;
        }
        return true;

    }

}
