/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.logging.Logger;

import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.text.BadLocationException;

import com.hcrest.jfreespace.inreport.HidInMsg;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutDataModeRequest;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutDataModeControlV2Request;
import com.hcrest.jfreespace.outreport.HidOutMsg;

/**
 * Example Swing application that exercises the jFreespace package. 
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
 * @version 1.0
 */
public class JFreespace extends JFrame implements DiscoveryListenerInterface, DeviceListenerInterface {
    private static final int MAX_LINE_COUNT = 500;
    private static final long serialVersionUID = 5547297743198823909L;
    private static final String DEVICE_STATISTICS = "No statistics";
    private Logger log_ = Logger.getLogger(JFreespace.class.getName());
    private JComboBox deviceSelection_;
    private JTextArea deviceText_;
    private JTextArea deviceStatistics_;
    private JTextArea reportText_;
    private JLabel    status_;
    private Device    device_ = null;  // The currently selected device.
    
    private Timer timer_ = null;
    private DeviceStatistics stats_ = null;


    /**
     * Create an instance of JFreespace.
     */
    public JFreespace() {
        super();
        
        instantiateComponents();

        Discovery discovery = Discovery.getInstance();
        discovery.addListener(this);
        
        timer_ = new Timer(1000, new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (device_ == null) {
                    return;
                }
                stats_ = device_.getStatistics(stats_);
                deviceStatistics_.setText(stats_.toString());
            }
        });
    }
    
    /**
     * Instantiate components.
     */
    private void instantiateComponents() {
        JPanel mainPanel = new javax.swing.JPanel();
        deviceSelection_ = new JComboBox();
        
        deviceText_ = new JTextArea("No device found");
        deviceText_.setRows(3);
        deviceText_.setEditable(false);
        deviceText_.setBorder(new TitledBorder(new EtchedBorder(), "Detailed Information"));
        deviceText_.setBackground(deviceSelection_.getBackground());
        
        deviceStatistics_ = new JTextArea(DEVICE_STATISTICS);
        deviceStatistics_.setRows(4);
        deviceStatistics_.setEditable(false);
        deviceStatistics_.setBorder(new TitledBorder(new EtchedBorder(), "Statistics"));
        deviceStatistics_.setBackground(deviceSelection_.getBackground());
        deviceStatistics_.setFont(new Font(Font.MONOSPACED, 
                deviceStatistics_.getFont().getStyle(), 
                deviceStatistics_.getFont().getSize()));
        
        reportText_ = new JTextArea("No incoming messages");
        reportText_.setEditable(false);
        reportText_.setBorder(new TitledBorder(new EtchedBorder(), "Reports"));
        reportText_.setBackground(deviceSelection_.getBackground());

        // Get the version information from the package
        Package p = this.getClass().getPackage();
        String ver = p.getSpecificationVersion();
        if (ver == null || ver.isEmpty()) {
            ver = "<unknown>";
        }
        JLabel version = new JLabel("Version = " + ver);
        
        status_ = new JLabel("Status: ");

        mainPanel.setLayout(new BorderLayout());

        JPanel selectionPanel = new JPanel();
        selectionPanel.setLayout(new FlowLayout());

        selectionPanel.add(new JLabel("Device"));
        deviceSelection_.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                updateDeviceSelection();
            }
        });
        selectionPanel.add(deviceSelection_, BorderLayout.NORTH);
        JPanel topPanel = new JPanel();
        topPanel.setLayout(new BorderLayout());
        
        topPanel.add(selectionPanel, BorderLayout.NORTH);
        topPanel.add(deviceText_, BorderLayout.SOUTH);
        mainPanel.add(deviceStatistics_, BorderLayout.NORTH);
        mainPanel.add(new JScrollPane(reportText_), BorderLayout.CENTER);

        getContentPane().removeAll();
        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(topPanel, BorderLayout.NORTH);
        getContentPane().add(mainPanel, java.awt.BorderLayout.CENTER);
        
        JPanel statusBar = new JPanel();
        statusBar.setLayout(new BorderLayout());
        statusBar.add(status_, BorderLayout.CENTER);
        statusBar.add(version, BorderLayout.EAST);
        getContentPane().add(statusBar, BorderLayout.SOUTH);
    }

    /**
     * Handle a user-signal application exit.
     * @param evt The exit window event.
     */
    private void exitForm(java.awt.event.WindowEvent evt) {
        timer_.stop();
        closeActiveDevice();
        Discovery discovery = Discovery.getInstance();
        discovery.removeListener(this);
    }
    
    /**
     * Display a demonstration GUI.
     * @param args The command line arguments (ignored).
     */
    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
            private JFreespace fs_;
            public void run() {
                fs_ = new JFreespace();
                fs_.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
                fs_.addWindowListener(new java.awt.event.WindowAdapter() {
                    @Override
                    public void windowClosing(java.awt.event.WindowEvent evt) {
                        fs_.exitForm(evt);
                    }
                });
                
                fs_.setResizable(true);
                fs_.setTitle("jFreespace");
                fs_.pack();
                fs_.setLocation(50, 50);
                fs_.setSize(new Dimension(600, 800));
                fs_.setVisible(true);
            }
        });
    }

    /**
     * Update the device selection to reflect the added or removed devices.
     */
    private void updateDeviceSelection() {
        Device newDevice = (Device) deviceSelection_.getSelectedItem();
        if (device_ == newDevice) {
            // Not really an update.
            return;
        }
        
        deviceText_.selectAll();
        deviceText_.replaceSelection("");
        
        reportText_.selectAll();
        reportText_.replaceSelection("");

        closeActiveDevice();
        stats_ = null;
        device_ = newDevice;
        if (device_ == null) {
            deviceText_.setText("No device found");
            return;
        }
        // Configure the new device.
        deviceText_.setText(device_.toDetailedString());
        log_.info("Open device: " + device_);
        timer_.restart();
        FreespaceErrorCodes ec = device_.open(this);
        if (ec != FreespaceErrorCodes.FREESPACE_SUCCESS) {
            log_.warning("Error " + ec.toString() + " while opening device " + device_);
        } else {
            log_.info("Device successfully opened");
        }
        
        /*
         * The following code is for FSP160 and FSP2XX devices that use the DataModeControl V1 message
         * 
         * Uncomment and comment out the next section if your device supports these devices
        
        FreespaceMsgOutDataModeRequest mr = new FreespaceMsgOutDataModeRequest();
        mr.setEnableBodyMotion(true);
        mr.setEnableUserPosition(true);
        mr.setInhibitPowerManager(true);        
        device_.sendMessageAsync(mr);
        
        */
        
        /*
         * The following code is for MotionEngine devices (Scoop, FSRK-USB-2)
         * 
         * Comment these and uncomment the block above for older devices
         */
        
        FreespaceMsgOutDataModeControlV2Request mr = new FreespaceMsgOutDataModeControlV2Request();
        mr.setModeAndStatus(8); // Always on Mode
        mr.setPacketSelect(2);  // Body Frame
        device_.sendMessageAsync(mr);
    }

    private void closeActiveDevice() {
        if (device_ != null) {
            // Close the existing device.
            FreespaceMsgOutDataModeRequest mr = new FreespaceMsgOutDataModeRequest();
            mr.setEnableMouseMovement(true);
            device_.sendMessage(mr);
            device_.close();
            timer_.stop();
        }
        deviceStatistics_.setText(DEVICE_STATISTICS);
        device_ = null;
        stats_ = null;
    }
    
    /**
     * Synchronize device inserted events to the Swing thread.
     * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
     * @version 1.0
     */
    private class FreespaceDeviceInserted implements Runnable {
        private final Device dev_;
        public FreespaceDeviceInserted(Device device) {
            dev_ = device;
        }
        /** {@inheritDoc} */
        @Override
        public void run() {
            deviceSelection_.addItem(dev_);
        }
    }
    
    /** {@inheritDoc} */
    @Override
    public void freespaceDeviceInserted(Device device) {
        SwingUtilities.invokeLater(new FreespaceDeviceInserted(device));
    }

    /**
     * Synchronize device removed events to the Swing thread.
     * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
     * @version 1.0
     */
    private class FreespaceDeviceRemoved implements Runnable {
        private final Device dev_;
        public FreespaceDeviceRemoved(Device device) {
            dev_ = device;
        }
        /** {@inheritDoc} */
        @Override
        public void run() {
            deviceSelection_.removeItem(dev_);
        }
    }
    
    /** {@inheritDoc} */
    @Override
    public void freespaceDeviceRemoved(Device device) {
        SwingUtilities.invokeLater(new FreespaceDeviceRemoved(device));
    }

    
    /**
     * Add the new text string to the reportText_ field.
     * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
     * @version 1.0
     */
    private class AddReportText implements Runnable {
        private final String str_;
        /**
         * Create a new instance of AddReportText.
         * @param str The string to add (append) to the text field.
         */
        public AddReportText(String str) {
            str_ = str;
        }

        /** {@inheritDoc} */
        @Override
        public void run() {
            int lc = reportText_.getLineCount(); 
            if (lc >= MAX_LINE_COUNT) {
                try {
                    reportText_.replaceRange("", 0, reportText_.getLineEndOffset(lc - MAX_LINE_COUNT));
                } catch (BadLocationException ex) {
                    // Critical error.
                }
            }
            reportText_.append(str_);
        }
    }

    /** {@inheritDoc} */
    public void handleRead(Device device, HidInMsg msg, int rc, long ts) {
        SwingUtilities.invokeLater(new AddReportText(msg.toDetailedString()));
    }

    /** {@inheritDoc} */
    @Override
    public void handleSend(Device device, HidOutMsg msg, int rc) {
        if (rc != 0) {
            log_.warning("Could not send message to " + device.toString());
        }
    }
    
    /** {@inheritDoc} */
    @Override
    public void notifyRemoved(Device device) {
        SwingUtilities.invokeLater(new AddReportText("Active device removed"));
    }

}
