/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URLConnection;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Extract files from a JAR into the appropriate directory.
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matthew Liberty</a>
 * @version 1.0
 */
public final class JarFileExtractor {
    private static final String BASE_DIR = "Hillcrest Labs";
    private static JarFileExtractor instance_ = null;
    private static String appDir_ = "JFreespace";
    private static String systemDir_ = null;
    private static Logger log_ = Logger.getLogger(JarFileExtractor.class.getName());
    private String dir_ = null;
    private Set<String> extractedFiles_ = new LinkedHashSet<String>();

    private JarFileExtractor() {
        super();
        findExtractionDirectory();
    }

    /**
     * Get the singleton instance.
     * @return The singleton instance.
     */
    public static JarFileExtractor getInstance() {
        if (instance_ == null) {
            instance_ = new JarFileExtractor();
        }
        return instance_;
    }
    
    /**
     * The name of the currently running application will be used as the 
     * subdirectory of BASE_DIR. The default is "Freespace Toolkit".
     * @param appDir A unique identifier per application.
     */
    public static void setAppDir(String appDir) {
        appDir_ = appDir;
    }

    /**
     * Get the application data directory.
     * @return The data data directory path name.
     */
    public static String getAppDataDir() {
        if (systemDir_ == null) {
            javax.swing.JFileChooser fr = new javax.swing.JFileChooser();
            systemDir_ = fr.getFileSystemView().getDefaultDirectory().toString();
        } 
        
        String userDir = systemDir_ +
            File.separator + BASE_DIR +
            File.separator + appDir_ + File.separator;
        
        if (!new File(userDir).exists()) {
            new File(userDir).mkdirs();
        }
        return userDir;
    }
    
    /**
     * Get the application data directory with the provided subdirectory.
     * @param subdir The subdirectory.
     * @return The data directory path name.
     */
    public static String getAppDataDir(String subdir) {
        if (subdir == null || subdir.isEmpty()) {
            return getAppDataDir();
        }
        String fulldir = getAppDataDir() + subdir + File.separator;
        if (!new File(fulldir).exists()) {
            new File(fulldir).mkdirs();
        }
        
        return fulldir;
    }
    
    /**
     * Identify the target extraction directory for the resources.
     */
    private void findExtractionDirectory() {
        // Get the user directory.
        String userDir = getAppDataDir() + "lib";

        // Create the full directory list in priority order
        String[] dirList = new String[] {
            "lib",
            userDir
        };

        // Find the first available directory.
        for (String dir : dirList) {
            try {
                File fDir = new File(dir);
                fDir.mkdirs();
                if (fDir.canWrite()) {
                    File test = new File(fDir, "test.out");
                    test.createNewFile();
                    test.delete();
                    
                    dir_ = fDir.getAbsolutePath();
                    log_.info("Target directory = " + dir);
                    return;
                }
            } catch (SecurityException se) {
                log_.warning("Target directory not write accessible: " + dir);
            } catch (IOException e) {
                log_.warning("Target directory not write accessible: " + dir);
            }
        }
        log_.warning("No target extraction directory found");
    }


    /**
     * Extract the desired files from the JAR using the srClass as reference.
     * @param srcClass The source object reference, or null to use global
     * resource identifier.
     * @param srcFile The desire source resource filenames.
     * @param dstDir The target destination directory.  Use null or "" to specify
     * the default extraction directory.
     * @return The extracted filenames on success, or a null value otherwise.
     */
    public String[] extract(Class<?> srcClass, String[] srcFile, String dstDir) {
        String[] strOut = new String[srcFile.length];
        for (int i = 0; i < srcFile.length; i++) {
            if (dstDir == null || dstDir.isEmpty()) {
                strOut[i] = extract(srcClass, srcFile[i], srcFile[i]);
            } else {
                strOut[i] = extract(srcClass, srcFile[i], dstDir + File.separator + srcFile[i]);
            }
        }
        return strOut;
    }


    /**
     * Extract the provided input stream to the destination file.
     * @param in The input stream.
     * @param dstFile The fully qualified filename.  The parent directory must exist.
     * @return True on success, false on failure.
     */
    public boolean extract(InputStream in, File dstFile) {
        // Copy the source resource file to the destination output file
        try {
            dstFile.createNewFile();

            FileOutputStream out = new FileOutputStream(dstFile);
            byte[] buffer = new byte[2048];
            int nBytes = 0;
            while ((nBytes = in.read(buffer)) >= 0) {
                out.write(buffer, 0, nBytes);
            }

            // Clean up after ourselves.
            out.flush();
            out.getFD().sync();
            out.close();
            in.close();
            out = null;
            dstFile = null;
            System.gc();
            System.gc();
        } catch (IOException ex) {
            if (!dstFile.exists()) {
                // We were not successful
                log_.log(Level.WARNING, "Could not extract file", ex);
                return false;
            } else {
                // Otherwise, just try to use the file that is already there.
                log_.log(Level.INFO, "Could not extract library, using existing version", ex);
            }
        }

        return true;
    }


    /**
     * Extract the desired file from the JAR using the srcClass as reference.
     * @param srcClass The source object reference, or null to use global
     * resource identifier.
     * @param srcFile The desired source resource filename.
     * @param dstFile The desired target resource filename with absolute path.
     * @return The extracted filename on success, null otherwise.
     */
    public String extractAbsolute(Class<?> srcClass, String srcFile, String dstFile) {
        File outFile = new File(dstFile);

        // Create the directory structure as needed.
        if (!outFile.getParentFile().exists() && !outFile.getParentFile().mkdirs()) {
            // Could not create the desired directory structure.
            return null;
        }

        // Open the input stream.
        InputStream in = null;
        long lastModified = System.currentTimeMillis();

        try {
            URLConnection connection;
            // Load as system or object relative resource as needed
            if (srcClass == null) {
                connection = ClassLoader.getSystemResource(srcFile).openConnection();
            } else {
                connection = srcClass.getResource(srcFile).openConnection();
            }
            lastModified = connection.getLastModified();
            in = connection.getInputStream();
        } catch (IOException ex) {
            log_.warning("Source file not found: " + srcFile);
            return null;
        }

        if (in == null) {
            log_.warning("Source file not found: " + srcFile);
            return null;
        }

        if (!extract(in, outFile)) {
            return null;
        }

        if (!(new File(dstFile)).exists()) {
            log_.info("Extracted but not yet present: " + dstFile);
        }

        outFile.setLastModified(lastModified);
        extractedFiles_.add(dstFile);
        return dstFile;
    }

    /**
     * Extract the desired file from the JAR using the srcClass as reference.
     * @param srcClass The source object reference, or null to use global
     * resource identifier.
     * @param srcFile The desired source resource filename.
     * @param dstFile The desired target resource filename.  This filename may contain
     * a directory structure that is relative to the default extraction location.
     * @return The extracted filename on success, null otherwise.
     */
    public String extract(Class<?> srcClass, String srcFile, String dstFile) {
        // Create the destination file.
        String outFilename = dir_ + File.separator + dstFile;
        return extractAbsolute(srcClass, srcFile, outFilename);
    }

    /**
     * Attempt to load a library that was extracted.  This method uses
     * <code>System.load</code> to load the library.  However, the load call
     * is surrounded by retries to allow the operating system to refresh.
     * With MS Windows, the library is not always immediately available
     * after being written to the extracted directory.
     *
     * @param libFilename The filename of the library to load.
     * @return True on success, false on failure.
     */
    public boolean loadLibrary(String libFilename) {
        for (int i = 0; i < 20; i++) {
            try {
                System.load(libFilename);
                log_.info("Library loaded: " + libFilename);
                return true;
            } catch (UnsatisfiedLinkError ex) {
                log_.info("Library failed load attempt " + i + " : " + libFilename);
                try {
                    Thread.sleep(50);
                } catch (InterruptedException e) {
                    // We do not care
                }
            }
        }
        log_.warning("Library could not be loaded: " + libFilename);
        return false;
    }
}
