/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;

/**
 * Supplies direct access to the libfreespace library.  External packages
 * should not access this class directly and should instead rely on 
 * com.hcrest.jfreespace.Discovery and com.hcrest.jfreespace.Device.
 * 
 * Below is the sequence of methods called to process an incoming message (i.e. USB HID IN report).
 * 
 * jfreespace.dll > jfreespace_receiveCallback     The last function in the dll that handles the message
 *   LibFreespace.readCallback                     Determines which device should get the message
 *     Device_instance.readCallback                Knows who is listening to this device
 *       listener_.handleRead                      Message is handed off to the listener
 *       
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
 * @version 1.0
 */
final class LibFreespace implements Runnable {
    private static final String LIBRARY = "jfreespace";
    private static final String LIBRARYX64 = "jfreespacex64";
    private static final int LIBFREESPACE_TIMEOUT = 1000; // in milliseconds
    private static boolean isLibraryLoaded_ = false;
    private static LibFreespace instance_ = null;
    private static Logger log_ = Logger.getLogger(LibFreespace.class.getName());

    private Collection<Device> devices_ = new CopyOnWriteArrayList<Device>();
    
    private ConcurrentLinkedQueue<FreespaceEvent> queue_ = new ConcurrentLinkedQueue<FreespaceEvent>();
    private Discovery discovery_ = null;
    private Thread thread_ = null;
    private FreespaceEvent initializeEvent_ = null;
    private int sendId_ = 0;
    private Map<Integer, FreespaceEvent> sendEvents_ = new HashMap<Integer, FreespaceEvent>();

    
    /**
     * Create the one and only one instance of LibFreespace.
     */
    private LibFreespace() {
        // No action required.
    }
    
    /**
     * Get the one and only one instance of LibFreespace.
     * @return The LibFreespace instance.
     */
    public static LibFreespace getInstance() {
        if (instance_ == null) {
            instance_ = new LibFreespace();
        }
        return instance_;
    }
    
    /**
     * Set the discovery instance observer.
     * @param discovery The observer.
     */
    void setDiscovery(Discovery discovery) {
        discovery_ = discovery;
    }
    
    /**
     * Initialize the native library on its own thread.
     */
    private void initialize() {
        if (thread_ != null) {
            return;
        }
        thread_ = new Thread(this);
        thread_.start();
    }
    
    /** {@inheritDoc} */
    @Override
    public void run() {
        if (!loadLibrary()) {
            initializeEvent_.setReturnCode(-1);
            return;
        }
        initializeEvent_.setReturnCode(native_init(this));
        log_.info("Start libfreespace event processing Loop");
        native_loop();
        log_.info("Stop libfreespace event processing Loop");
    }
    
    /**
     * Send a command event to the native library.  This method automatically 
     * synchronizes to the LibFreespace thread and is thread safe.
     * @param event The command event.
     * @return True if command processed, false if still pending.
     */
    static boolean sendEvent(FreespaceEvent event) {
        LibFreespace instance = getInstance();
        if (event.getType() == FreespaceEvent.Type.INITIALIZE) {
            instance.initializeEvent_ = event;
            instance.initialize();
            return false;
        }
        
        if (Thread.currentThread() == instance.thread_) {
            // Already on the correct thread!  Just perform the event.
            instance.performEvent(event);
            return true;
        } else {
            // Queue the event for handling on the libfreespace thread.
            instance.queue_.add(event);
            native_signal();
            return false;
        }
    }
    
    /**
     * Send a command event to the native library and wait for it to complete.
     * This method automatically synchronizes to the LibFreespace thread and 
     * is thread safe.
     * @param event The command event.
     * @return The return code.
     */
    static int sendEventSynchronized(FreespaceEvent event) {
        if (!sendEvent(event)) {
            // Event has been queued.  Wait for the response.
            if (!event.waitBoolean(LIBFREESPACE_TIMEOUT)) {
                log_.warning("Timed out waiting for event to complete.");
                return -1;
            }
        }
        return event.getReturnCode();
    }
    
    /**
     * Perform any queued FreespaceEvents.
     * @return The number of events processed, or a negative number on error.
     */
    int perform() {
        int count = 0;
        while (true) { 
            FreespaceEvent ev = queue_.poll();
            if (ev == null) {
                return count;
            }
            
            int rc = performEvent(ev);
            if (rc < 0) {
                return rc;
            } else {
                count += rc;
            }
        }
    }
    
    /**
     * Perform the FreespaceEvent.
     * @param ev The Freespace event to perform.
     * @return 1 if the event was processed, 0 if the event was not processed,
     *  or a negative number on error.
     */
    private int performEvent(FreespaceEvent ev) {
        switch (ev.getType()) {
            case INITIALIZE:
                // Should be handled by run().
                break;
                
            case EXIT:
                native_exit();
                ev.setReturnCode(0);
                return -1;
                
            case OPEN:
                ev.setReturnCode(native_openDevice(ev.getDeviceId()));
                break;
                
            case CLOSE:
                native_closeDevice(ev.getDeviceId());
                ev.setReturnCode(0);
                break;

            case FLUSH:
                ev.setReturnCode(native_flush(ev.getDeviceId()));
                break;
                
            case SEND:
                // Add message to send queue.
                sendId_++;
                sendEvents_.put(sendId_, ev);
                int rc = native_send(ev.getDeviceId(), ev.getMessageBytes(), sendId_);
                if (rc != 0) {
                    sendCallback(ev.getDeviceId(), rc, sendId_);
                }
                break;
                
            default:
                return 0;
        }
        return 1;
    }
    

    /**
     * Load the supporting library from the class (jar) file.
     * @return True on library successfully loaded, false on error.
     */
    private boolean loadLibrary() {
        if (isLibraryLoaded_) {
            return isLibraryLoaded_;
        }
        
        String libname;
        
        // Find the library as a resource file
        // Differentiate between native architecture, since the 32-bit VM cannot load
        //   64-bit native code and vice versa.
        try {
            String archModel = java.lang.System.getProperty("sun.arch.data.model");
            
            if (archModel.equals("64")) {
                // Use 64-bit library.
                libname = System.mapLibraryName(LIBRARYX64);
                log_.info("Using x64");
            } else {
                // Use 32-bit library.
                libname = System.mapLibraryName(LIBRARY);
                log_.info("Using Win32");
            }
        } catch (SecurityException e) {
            log_.severe("Security exception trying to determine 32 or 64 bit.  Defaulting to 32-bit.");
            
            // Default to 32-bit library.
            libname = System.mapLibraryName(LIBRARY);
        } catch (NullPointerException e) {
            log_.severe("Null pointer exception trying to determine 32 or 64 bit.  Defaulting to 32-bit.");
            
            // Default to 32-bit library.
            libname = System.mapLibraryName(LIBRARY);
        } catch (IllegalArgumentException e) {
            log_.severe("Illegal argument exception trying to determine 32 or 64 bit.  Defaulting to 32-bit.");
            
            // Default to 32-bit library.
            libname = System.mapLibraryName(LIBRARY);
        }
       
        // Determine the output library absolute path
        JarFileExtractor jfx = JarFileExtractor.getInstance();
        String libFilename = jfx.extract(LibFreespace.class, libname, libname);
        isLibraryLoaded_ = jfx.loadLibrary(libFilename);
        
        return isLibraryLoaded_;
    }

    /**
     * Get the device by its ID.
     * @param id The ID of the desired device.
     * @return The device, or null if no matching device was found.
     */
    private Device getDeviceById(int id) {
        for (Device d : devices_) {
            if (d.getId() == id) {
                return d;
            }
        }
        return null;
    }
    
    /**
     * Callback from libfreespace on device inserted.
     * @param device The inserted device.
     */
    void deviceInserted(Device device) {
        Device lastDevice = getDeviceById(device.getId());
        if (lastDevice != null) {
            log_.warning("Device ID reuse detected: " + device.getId());
            return;
        }
        
        log_.info("Adding device: " + device);
        devices_.add(device);
        
        if (discovery_ != null) {
            discovery_.deviceInserted(device);
        }
    }
    
    /**
     * Callback from libfreespace on device removed.
     * @param id The removed device id.
     */
    void deviceRemoved(int id) {
        Device device = getDeviceById(id);
        if (device == null) {
            log_.warning("Removing an unknown device: " + id);
            return;
        }
        devices_.remove(device);
        log_.info("Removing device: " + device);
        device.removed();

        if (discovery_ != null) {
            discovery_.deviceRemoved(device);
        }
    }
    
    /**
     * Get all currently connected devices.
     * @return The collection of devices.
     */
    Collection<Device> getDevices() {
        return devices_;
    }
    
    /**
     * Read callback from libfreespace on read message received.
     * @param The device ID.
     * @param msg The message received.
     * @param rc The return code from the read.
     * @param ts A relative timestamp indicating when the message was received.
     */
    void readCallback(int id, byte[] msg, int rc, long ts) {
        Device device = getDeviceById(id);
        if (device != null) {
            if (rc == 0) {
                device.readCallback(msg, rc, ts);
            } else {
               log_.warning("Read error from libfreespace: " + rc); 
            }
        }
    }
    
    /**
     * Send callback from libfreespace on send message completed.
     * @param id The device ID.
     * @param errorCode The return code from the send command.
     * @param cookie The send event ID.
     */
    void sendCallback(int id, int errorCode, int cookie) {
        FreespaceEvent ev = sendEvents_.remove(cookie);
        if (ev == null) {
            return;
        }
        ev.setReturnCode(errorCode);
        
        Device device = getDeviceById(id);
        if (device != null) {
            device.sendCallback(ev);
        }
    }

    private static native int native_init(LibFreespace lib);
    private static native void native_exit();
    private static native int native_openDevice(int id);
    private static native void native_closeDevice(int id);
    private static native int native_send(int id, byte[] msg, int cookie);
    private static native int native_flush(int id);
    private static native void native_signal();
    private static native void native_loop();
    
    /** {@inheritDoc} */
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("LibFreespace singleton instance:\n");
        buf.append("   " + devices_.size() + " devices connected\n");
        buf.append("   " + sendEvents_.size() + " outstanding send events\n");
        buf.append("   " + queue_.size() + " outstanding command events\n");
        return buf.toString();
    }
}
