/*
 * This file is part of jfreespace.
 *
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc.
 *
 * jfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import com.hcrest.jfreespace.inreport.HidInMsg;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutBatteryLevelRequest;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutDataModeRequest;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutProductIDRequest;
import com.hcrest.jfreespace.outreport.HidOutMsg;

/**
 * Get continuous data from every Freespace device plugged into the system.
 * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
 * @version 1.0
 */
public class MultiDeviceExample implements DiscoveryListenerInterface, DeviceListenerInterface {
    private final List<Device> devices_ = new CopyOnWriteArrayList<Device>();
    private Logger log_ = Logger.getLogger(MultiDeviceExample.class.getName());
    private static final int DEVICE_STATUS_ROWS = 18;
    private static final int DEVICE_STATUS_COLUMNS = 3;
    private DeviceStatus[] deviceStatus_ = new DeviceStatus[DEVICE_STATUS_ROWS * DEVICE_STATUS_COLUMNS];
    private JFrame frame_ = null;
    
    private final Timer timer_;

    /**
     * Create an instance of MultiDeviceExample.
     */
    public MultiDeviceExample() {
        Discovery discovery = Discovery.getInstance();
        discovery.addListener(this);
        
        timer_ = new Timer("MultiDeviceExample-Timer");
        TimerTask task = new TimerTask() {
            int tickCount_ = 0;
            @Override
            public void run() {
                tickCount_++;
                if (tickCount_ >= 10) {
                    getStats();
                    sendMsg();
                    tickCount_ = 0;
                } else {
                    sendMsg();
                }
            }
            private void getStats() { 
                int idx = 0;
                for (Device device : devices_) {
                    DeviceStatistics s = device.getStatistics();
                    deviceStatus_[idx].update(device.getId(), 
                            (int) s.getSendMessageCount(),
                            (int) s.getReadMessageCount());

                    idx++;
                }
                // Clear the remaining fields.
                for (int index = idx; index < DEVICE_STATUS_ROWS * DEVICE_STATUS_COLUMNS; index++) {
                    deviceStatus_[index].update(-1, 0, 0);
                }
            }
            
            private void sendMsg() {
                HidOutMsg msg;
                for (Device device : devices_) {
                    // Send a battery level request
                    msg = new FreespaceMsgOutBatteryLevelRequest();
                    device.sendMessageAsync(msg);
                    // Send a product ID request
                    msg = new FreespaceMsgOutProductIDRequest();
                    switch (device.getHidVer()) {
                        case FSP16x:
                            // The RFT will forward the request to the remote
                            device.sendMessageAsync(msg);
                            break;
                        case FSP2xx:
                            // Send a request to the RFT
                            msg.setDest(1);
                            device.sendMessageAsync(msg);
                            // Send a request to the remote
                            msg.setDest(4);
                            device.sendMessageAsync(msg);
                            break;
                        default:
                            device.sendMessageAsync(msg);
                            break;
                    }
                }
            }
        };
        timer_.scheduleAtFixedRate(task, 100, 200);
    }
    
    /**
     * Stop execution for this example.
     */
    public void stop() {
        timer_.cancel();
        for (Device device : devices_) {
            freespaceDeviceRemoved(device);
        }
        Discovery discovery = Discovery.getInstance();
        discovery.removeListener(this);
    }
    
    
    /** {@inheritDoc} */
    @Override
    public void freespaceDeviceInserted(Device device) {
        log_.info("Inserted device : " + device);
        FreespaceErrorCodes ec = device.open(this);
        if (ec != FreespaceErrorCodes.FREESPACE_SUCCESS) {
            log_.warning("Error " + ec.toString() + " while opening device " + device);
            return;
        }
        FreespaceMsgOutDataModeRequest msg = new FreespaceMsgOutDataModeRequest();
        msg.setDest(4);
        msg.setEnableBodyMotion(true);
        device.sendMessageAsync(msg);
        devices_.add(device);
    }

    /** {@inheritDoc} */
    @Override
    public void freespaceDeviceRemoved(Device device) {
        FreespaceMsgOutDataModeRequest msg = new FreespaceMsgOutDataModeRequest();
        msg.setDest(4);
        msg.setEnableMouseMovement(true);
        device.sendMessageAsync(msg);
        devices_.remove(device);
        device.close();
    }

    /** {@inheritDoc} */
    @Override
    public void handleRead(Device device, HidInMsg msg, int rc, long ts) {
        if (rc != 0) {
            log_.warning("Error code " + rc + " on read from " + device);
        }
    }

    /** {@inheritDoc} */
    @Override
    public void handleSend(Device device, HidOutMsg msg, int rc) {
        if (rc != 0) {
            log_.warning("Error code " + rc + " on send to " + device);
        }
    }

    /** {@inheritDoc} */
    @Override
    public void notifyRemoved(Device device) {
        // No action required.
    }
    
    /**
     * The device status for a single device.
     * @author <a href="mailto:Matt.Liberty@Hillcrestlabs.com">Matt Liberty</a>
     * @version 1.0
     */
    private class DeviceStatus {
        private final JLabel label_;
        private final JLabel out_;
        private final JLabel in_;
        
        /**
         * Create an instance of DeviceStatus.
         */
        public DeviceStatus() {
            label_ = new JLabel("_____ ");
            out_   = new JLabel("_____ ");
            in_    = new JLabel("_____ ");
            Font boldFont = out_.getFont().deriveFont(Font.BOLD);
            Font plainFont = out_.getFont().deriveFont(Font.PLAIN);
            label_.setFont(boldFont);
            out_.setFont(plainFont);
            in_.setFont(plainFont);
        }
        
        /**
         * Add the GUI for this status to a component.
         * @param component The parent component.
         */
        public void add(JComponent component) {
            component.add(label_);
            component.add(out_);
            component.add(in_);
        }
        
        /**
         * Update the GUI component.
         * @param id The device ID.
         * @param out The number of out (send) reports.
         * @param in The number of in (read) reports.
         */
        public void update(int id, int out, int in) {
            if (id < 0) {
                label_.setText(" ");
                out_.setText(" ");
                in_.setText(" ");
            } else {
                label_.setText(Integer.toString(id));
                out_.setText(Integer.toString(out));
                in_.setText(Integer.toString(in));
            }
        }
        
    }
    
    /**
     * Construct and open the GUI frame.
     */
    private void openFrame() {
        frame_ = new JFrame();
        frame_.setTitle("MultiDevice");
        frame_.setVisible(true);
        frame_.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        frame_.addWindowListener(new java.awt.event.WindowAdapter() {
            @Override
            public void windowClosing(java.awt.event.WindowEvent evt) {
                stop();
            }
        });
        
        JPanel top = new JPanel();
        top.setLayout(new BorderLayout());
        top.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
        JLabel label = new JLabel("id  send  read");
        top.add(label, BorderLayout.NORTH);
        JPanel panel = new JPanel();
        panel.setLayout(new GridLayout(DEVICE_STATUS_ROWS, 3 * DEVICE_STATUS_COLUMNS));
        
        for (int idx = 0; idx < DEVICE_STATUS_ROWS * DEVICE_STATUS_COLUMNS; idx++) {
            deviceStatus_[idx] = new DeviceStatus();
            deviceStatus_[idx].add(panel);
        }
        top.add(panel, BorderLayout.CENTER);
        frame_.setContentPane(top);
        frame_.pack();
    }
    
    /**
     * Display a demonstration GUI.
     * @param args The command line arguments (ignored).
     */
    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                MultiDeviceExample ex = new MultiDeviceExample();
                ex.openFrame();
            }
        });
    }
}
