/*
 * Copyright 2009-2012 Hillcrest Laboratories, Inc. All rights reserved.
 * HILLCREST PROPRIETARY/CONFIDENTIAL.
 *
 * Reviewed:
 *   Never
 */
package com.hcrest.jfreespace;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Logger;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

import com.hcrest.jfreespace.inreport.HidInMsg;
import com.hcrest.jfreespace.inreport.HidInMsgType;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutBatteryLevelRequest;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutDataModeRequest;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutLinkQualityRequest;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutPairingMessage;
import com.hcrest.jfreespace.outreport.FreespaceMsgOutProductIDRequest;
import com.hcrest.jfreespace.outreport.HidOutMsg;

/**
 * A simple test to exercise some of the HIS messages.
 * @author <a href="mailto:Mark.Turner@Hillcrestlabs.com">Mark Turner</a>
 * @version 1.0
 */
public class TestHidMessages extends JFrame implements DiscoveryListenerInterface, DeviceListenerInterface {
    
    private static final long serialVersionUID = 5547297988274313909L;
    private Device    device_ = null;  // The currently selected device.
    private Logger log_ = Logger.getLogger(TestHidMessages.class.getName());
    
    private JPanel ctrlPanel_;
    private JButton runButton_;
    private JTextArea results_;
    private JLabel deviceInfoLabel_;
    private static final String NOT_AVAILABLE = "Not Available";
    
    private static long TIMEOUT = 5;
    private Timer testTimer_;
    private long tickCount_ = 0;
    private long sendTickStamp_ = 0;
    
    private enum TestState {
        IDLE,
        RUNNING
    }
    private int testCaseIdx_ = 0;
    
    private TestState testState_ = TestState.IDLE;
    
    private enum TestCase {
        PROD_ID(new FreespaceMsgOutProductIDRequest(), HidInMsgType.FREESPACE_MESSAGE_PRODUCTIDRESPONSE),
        BATTERY(new FreespaceMsgOutBatteryLevelRequest(), HidInMsgType.FREESPACE_MESSAGE_BATTERYLEVEL),
        DATA_MODE(new FreespaceMsgOutDataModeRequest(), HidInMsgType.FREESPACE_MESSAGE_DATAMODERESPONSE),
        PAIRING(new FreespaceMsgOutPairingMessage(), HidInMsgType.FREESPACE_MESSAGE_PAIRINGRESPONSE),
        LINK_STATUS(new FreespaceMsgOutLinkQualityRequest(), HidInMsgType.FREESPACE_MESSAGE_LINKSTATUS),
        ;
        
        private HidOutMsg outMsg_;
        private HidInMsgType inType_;
        
        private TestCase(HidOutMsg outMsg, HidInMsgType inType) {
            outMsg_ = outMsg;
            inType_ = inType;
        }
        
        private HidOutMsg getOutMsg() {
            return outMsg_;
        }
        
        private HidInMsgType getInType() {
            return inType_;
        }
    }
    
    /**
     * Create an instance of TestHidMessages.
     */
    public TestHidMessages() {
        super();
        
        instantiateComponents();

        Discovery discovery = Discovery.getInstance();
        discovery.addListener(this);

        // Set up a periodic timer to generate ticks during the test
        testTimer_ = new Timer("TestTimerThread");
        TickTask t = new TickTask();
        t.setTester_(this);
        testTimer_.scheduleAtFixedRate(t, 100, 100);
    }
    
    /**
     * Instantiate components.
     */
    private void instantiateComponents() {
        JPanel mainPanel = new javax.swing.JPanel();
        mainPanel.setBorder(new TitledBorder(new EtchedBorder(), "Simple HID Message Test"));
        mainPanel.setLayout(new BorderLayout());
        ctrlPanel_ = new JPanel();
        ctrlPanel_.setLayout(new BorderLayout());
        deviceInfoLabel_ = new JLabel(NOT_AVAILABLE);
        deviceInfoLabel_.setBorder(new TitledBorder(new EtchedBorder(), "Device Information"));
        runButton_ = new JButton("Run");
        runButton_.addActionListener(new RunButtonActionListener());
        runButton_.setEnabled(false);
        ctrlPanel_.add(deviceInfoLabel_, BorderLayout.NORTH);
        ctrlPanel_.add(runButton_, BorderLayout.SOUTH);
        
        results_ = new JTextArea();
        JScrollPane resultsScroll = new JScrollPane(results_);
        resultsScroll.setBorder(new TitledBorder(new EtchedBorder(), "Reports"));
        results_.setBackground(mainPanel.getBackground());
        mainPanel.add(ctrlPanel_, BorderLayout.NORTH);
        mainPanel.add(resultsScroll, BorderLayout.CENTER);
        getContentPane().add(mainPanel, java.awt.BorderLayout.CENTER);
    }
    

    /**
     * Handle a user-signal application exit.
     * @param evt The exit window event.
     */
    private void exitForm(java.awt.event.WindowEvent evt) {
        if (device_ != null) {
            // Close the existing device.
            device_.close();
            device_ = null;
        }
        Discovery discovery = Discovery.getInstance();
        discovery.removeListener(this);
        testTimer_.cancel();
    }
    
    /**
     * Handle the Run button.
     */
    private class RunButtonActionListener implements ActionListener {
        public void actionPerformed(ActionEvent evt) {
            String cmd = evt.getActionCommand();
            if (cmd.equals("Run")) {
                testCaseIdx_ = 0;
                runTestCase();
                testState_ = TestState.RUNNING;
                runButton_.setEnabled(false);
            }
        }
    }
    
    /**
     * Runs a simple test over some of the HID messages.
     * @param args Unused.
     */
    public static void main(String[] args) {
        final TestHidMessages fs;
        fs = new TestHidMessages();
        fs.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        fs.addWindowListener(new java.awt.event.WindowAdapter() {
            @Override
            public void windowClosing(java.awt.event.WindowEvent evt) {
                fs.exitForm(evt);
            }
        });
        
        fs.setResizable(true);
        fs.setTitle("JFreespace Test HID Messages");
        fs.pack();
        fs.setLocation(50, 50);
        fs.setSize(new Dimension(600, 800));
        fs.setVisible(true);
    }

    /** {@inheritDoc} */
    @Override
    public void handleRead(Device device, HidInMsg msg, int rc, long ts) {
        if (testState_ == TestState.RUNNING) {
            if (msg.getMsgType() == TestCase.values()[testCaseIdx_].getInType()) {
                results_.append("Pass\n");
                testCaseIdx_++;
                runTestCase();
            } else {
                results_.append("Unexpected message.\n");
            }
        }
    }

    /** {@inheritDoc} */
    @Override
    public void handleSend(Device device, HidOutMsg msg, int rc) {
        // NO action required. Is this method ever needed?
        
    }

    /** {@inheritDoc} */
    @Override
    public void notifyRemoved(Device device) {
        if (device_ == device) {
            // Close the existing device.
            device_.close();
            deviceInfoLabel_.setText(NOT_AVAILABLE);
            runButton_.setEnabled(false);
        }
        device_ = null;
    }

    /** {@inheritDoc} */
    @Override
    public void freespaceDeviceInserted(Device device) {
        if (device_ == null) {
            device_ = device;
            FreespaceErrorCodes ec = device_.open(this);
            if (ec != FreespaceErrorCodes.FREESPACE_SUCCESS) {
                log_.warning("Error " + ec.toString() + " while opening device " + device_);
            } else {
                log_.info("Device successfully opened");
                deviceInfoLabel_.setText(device_.toString());
                runButton_.setEnabled(true);
            }
        }
    }

    /** {@inheritDoc} */
    @Override
    public void freespaceDeviceRemoved(Device device) {
        if (device_ != device) {
            // Close the existing device.
            device_.close(); // TODO will removing a second device cause the first device to be closed?
            device_ = null;
            deviceInfoLabel_.setText(NOT_AVAILABLE);
        }
    }
    
    private void runTestCase() {
        if (testCaseIdx_ < TestCase.values().length) {
            HidOutMsg msg = TestCase.values()[testCaseIdx_].getOutMsg();
            msg.setDest(4);
            device_.sendMessage(msg);
            results_.append(msg.getMsgType().toString() + "\n");
            sendTickStamp_ = tickCount_;
        } else {
            results_.append("Test complete.\n");
            testState_ = TestState.IDLE;
            runButton_.setEnabled(true);
        }
    }

    private void timerTick() {
        tickCount_++;
        if (testState_ == TestState.RUNNING) {
            if (tickCount_ - sendTickStamp_ > TIMEOUT) {
                testCaseIdx_++;
                results_.append("TIMEOUT\n");
                runTestCase();
            }
        }
    }
    
    private class TickTask extends TimerTask {
        private TestHidMessages tester_;

        @Override
        public void run() {
            tester_.timerTick();
        }

        /**
         * Set the value of tester.
         * @param tester The tester to set.
         */
        public void setTester_(TestHidMessages tester) {
            tester_ = tester;
        }
        
        
    }
}
