/*
 * This file is part of libfreespace.
 * 
 * Copyright (c) 2009-2012 Hillcrest Laboratories, Inc. 
 *
 * libfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.hcrest.jfreespace.inreport;

/**
 * DecoderFactory.
 * Produces a HID IN message decoder.
 * @version 1.0
 */
final public class DecoderFactory {

    public enum HID_MESSAGE_VERSION {
        /** BOOTLOADER. */
        BOOTLOADER,
        /** FSP16x. */
        FSP16x,
        /** FSP2xx. */
        FSP2xx
    }
    HID_MESSAGE_VERSION t = HID_MESSAGE_VERSION.values()[2];
    
    /**
     * getDecoder returns the correct decoder to use for a given HID message version.
     * Use HID_MESSAGE_VERSION.values()[<ver>] where <ver> is retrieved from the device
     * to map the version number to the enumerated version type.
     * @param version The version of the HID messages being used.
     * @return
     */
    static public Decoder getDecoder(HID_MESSAGE_VERSION version) {
        Decoder d;
        switch (version) {
            case BOOTLOADER:
                d = DecoderV0.getInstance();
                return d;
            case FSP16x:
                d = DecoderV1.getInstance();
                d.setSubDecoder(SubDecoderV1.getInstance());
                return d;
            case FSP2xx:
                d = DecoderV2.getInstance();
                d.setSubDecoder(SubDecoderV2.getInstance());
                return d;
            default:
                throw new RuntimeException("Unknown version number.");
        }
    }
    
    /**
     * byteArrayToInt.
     * Extracts bytes from a byte array and converts them to a signed integer.
     * The LSB is in the lower array index and the MSB is in the upper array index.
     * If size is 1 or 2 then the byte(s) will be signed extended
     * @param b The byte array
     * @param offset Index of the LSB within the byte array
     * @param size How bytes to convert
     * @return A signed integer produce from the bytes in the byte array
     */
    static int byteArrayToInt(byte [] b, int offset, int size) {
        int si = b[offset + size - 1]; // Assign the MSB first so the sign extension happens
        for (int i = offset + size - 2; i >= offset; --i) { // Handle remaining bytes
            si = (si << 8) + (b[i] & 0xff); // Mask off sign extension from bytes that are not the MSB
        }
        return si;
    }
    
    /**
     * byteArrayToUnsignedInt.
     * Extracts bytes from a byte array and converts them to an unsigned integer.
     * The LSB is in the lower array index and the MSB is in the upper array index.
     * Okay, in Java it's still really a signed integer. But, if you convert 1, 2 or 3 bytes
     * and the MS bit in the MS byte is set, you will get a positive number so it looks like
     * an unsigned conversion.
     * @param b The byte array
     * @param offset Index of the LSB within the byte array
     * @param size How bytes to convert
     * @return A signed integer produce from the bytes in the byte array
     */
    static int byteArrayToUnsignedInt(byte [] b, int offset, int size) {
        int si = b[offset + size - 1] & 0xff; // Assign the MSB and mask off any sign extension
        for (int i = offset + size - 2; i >= offset; --i) { // Handle remaining bytes
            si = (si << 8) + (b[i] & 0xff); // Mask off sign extension
        }
        return si;
    }
    
}
